/* -*- c-file-style: "linux"; -*- */
/**********************************************************************
 * connection_table.c                                       August 2001
 * Horms                                             horms@vergenet.net
 *
 * Manage connections
 *
 * iproxy
 * UDP - TCP proxy
 * Copyright (C) 2001  Andrew Tridgell <tridge@samba.org>
 *                     Horms <horms@vergenet.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 **********************************************************************/

#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <stdlib.h>

#include "util.h"
#include "iproxy.h"
#include "packet.h"
#include "interface.h"
#include "connection.h"
#include "connection_table.h"


/**********************************************************************
 * Static table of connections.
 *
 * I seem to be using a lot of tables at the moment.
 * Tables are my friends...
 *
 * And all my friends are fuck-ups              
 * But they're fun to have around
 * Banana chairs out on the concrete
 * Telling stories to the stars
 * How Geminis love Wooden Dragons
 * How down the street the weeks roll by
 * Weeks roll by
 *     From "You Sound Like Louis Burdett", by The Whitlams
 **********************************************************************/

static struct connection con_table[MAX_CONNECTION];
static struct in_addr xcast_addr;


/**********************************************************************
 * connection_table_init
 * Initialise the connection table
 * pre: dest_ip: Address used for replies over UDP portion of connections.
 *               Ignored in Unicast mode, the IP address of the peer is
 *               used instead.
 * post: Connection table is intialised
 *       Note: This will remove any knowledge of existing connections.
 * return: 0 on success
 *         -1 on error
 **********************************************************************/

int connection_table_init(const char *dest_ip)
{
        memset(con_table, 0, sizeof(con_table));

        if (inet_pton(AF_INET, dest_ip, &xcast_addr) <= 0) {
                perror("connection_table_init: inet_pton");
                return(-1);
        }

        return(0);
}


/**********************************************************************
 * connection_table_dump
 * Dump information about each non-PURGED connection in the connection
 * table using(connection_dump)
 * pre: none
 * post: Nothing if connection_table_init hasn't been called
 *       Information about each connection is printed to stdout otherwise
 * return: none
 **********************************************************************/

void connection_table_dump(void)
{
        int con;
        char *tmp_str;
        time_t time_tmp;
        int active=0;

        time_tmp=time(NULL);
        tmp_str=ctime(&time_tmp);
        *(tmp_str+strlen(tmp_str)-1)='\0';
        printf("Dumping connection table %s: ", tmp_str);
        for(con=0;con<MAX_CONNECTION;con++){
                if(connection_table_get_status(con)==STATUS_PURGED)
                        continue;
                active++;
                printf("\nconnection: %d\n", con);
                connection_dump(&(con_table[con]));
                printf("\n");
        }

        printf("Done %d of %d active\n", active, MAX_CONNECTION);
        fflush(stdout);
}


/**********************************************************************
 * connection_table_get_connection
 * Find the connection associated with a packet
 * If the packet is a command packet, process it accodingly
 * If the packet is for an unknown connection, and isn't a command
 * packet, create a connection for it.
 * pre: dest_ip: Destination IP for outgoing TCP sockets
 *      dest_port: Destination Port for outgoing TCP sockets
 *      udp_peer: sockaddr_in (IP address and port) of the peer
 *                that sent the packet
 *      packet: the packet that was sent
 *      len: the length of the packet
 * post: If the packet is a command packet
 *          Process using connection_command_packet()
 *          Return
 *       Else If the connection is unknown
 *          Create a new connection
 *       Return connection
 * return: Connection number
 *         MAX_CONNECTION if connection cannot be found
 *                        or if an error occured handling a command packet
 *                        for a known connection
 *         MAX_CONNECTION+1 if an error occured handling a command packet
 *                          for an unknown connection
 *         MAX_CONNECTION+2 If a command packet was handled successfully
 *         MAX_CONNECTION+3 if it is a packet for a similar connection
 *         That is we are reciving duplicate packets, most likely
 *         because the host has multiple interfaces routed to the other
 *         host.
 **********************************************************************/

int connection_table_get_connection(char *dest_ip, unsigned short dest_port,
                struct sockaddr_in *udp_peer, struct packet *packet, 
                int length)
{
        int s;
        int con;
        int status;
        int oldest_con;
        time_t expire=0;

        extern int iproxy_verbose;
        extern uint32 iproxy_mode;

        for(con=0;con<MAX_CONNECTION;con++){
                if(con_table[con].udp_peer_port==udp_peer->sin_port && 
                   con_table[con].cid==ntohl(packet->cid) && 
                   con_table[con].udp_peer_addr.s_addr==
                                udp_peer->sin_addr.s_addr){
                        if(iproxy_verbose > 2)
                                printf("found connection %d in table\n", con);
                        if(length==PACKET_HDR_LEN && ntohl(packet->flag)){
                                goto command_packet;
                        }
                        return(con);
                }
        }

        /* If we get this far then the connection is not known */

        /* Throw away if we already have a similar connection */
        for(con=0;con<MAX_CONNECTION;con++){
                if(con_table[con].udp_peer_port==udp_peer->sin_port && 
                   con_table[con].cid==ntohl(packet->cid)){
                        if(iproxy_verbose > 2)
                                printf("found similar connection %d in "
                                                       "table\n", con);
                        return(MAX_CONNECTION+3);
                }
        }


        /* Handle command packets, then don't need a connection */
        if(length==PACKET_HDR_LEN && ntohl(packet->flag)){
                goto command_packet;
        }

        /* If we get this far then we have a new connection
         * find a free slot */
        for(con=0;con<MAX_CONNECTION;con++){
                if(con_table[con].status==STATUS_PURGED){
                        break;
                }
        }

        /* If there are no free connections, get the oldest closed
         * (but un-purged connection) */
        if(con==MAX_CONNECTION){
                oldest_con=MAX_CONNECTION;
                for(con=0;con<MAX_CONNECTION;con++) {
                        if(connection_table_get_status(con)!=STATUS_CLOSED)
                                continue;
                        if(oldest_con==MAX_CONNECTION || 
                           connection_table_get_expire(con)<expire){
                                oldest_con=con;
                                expire=connection_table_get_expire(con);
                        }
                }
                con=oldest_con;
                connection_table_purge(con);
        }

        if(con==MAX_CONNECTION){
                fprintf(stderr, "No free connections\n");
                if(!(iproxy_mode&MODE_UNICAST))
                        udp_peer->sin_addr=xcast_addr;
                return(MAX_CONNECTION);
        }

        /* Don't open TCP socket for a list request */
        if(ntohl(packet->sid) == IPROXY_LIST_SID){
                s=-1;
                goto initialise;
        }

        /* Open outgoing TCP socket */
        s=open_socket_out(SOCK_STREAM, dest_ip, dest_port, 0);
        if (s==-1) {
                fprintf(stderr, "Can't contact remote server!\n");
                memset(&con_table[con], 0, sizeof(con_table[con]));
                if(!(iproxy_mode&MODE_UNICAST))
                        udp_peer->sin_addr=xcast_addr;
                packet_send_end(udp_peer, htonl(packet->cid),
                                        htonl(packet->sid));
                return(MAX_CONNECTION);
        }

initialise:
        connection_initialise(&(con_table[con]), s, udp_peer, &xcast_addr,
                        ntohl(packet->cid), ntohl(packet->sid));
        
        return(con);

command_packet:
        status=connection_command_packet(
                        (con==MAX_CONNECTION)?NULL:&(con_table[con]),
                        packet, 0);
                        
        if(status==-1) {
                fprintf(stderr, "connection_table_get_connection: "
                                "connection_command_packet\n");
                return(MAX_CONNECTION+1);
        }

        if(status < 0) {
                fprintf(stderr, "connection_table_get_connection: "
                                "connection_command_packet: %d\n", con);
                return(MAX_CONNECTION);
        }

        if(status == 1){
                if(iproxy_verbose > 3)
                        printf("removing connection %d from table\n", con);
        }

        return(MAX_CONNECTION+2);
}


/**********************************************************************
 * connection_table_read_tcp
 * Read pending data from the TCP socket associated with each
 * un-PURGED connection and pass it to the corresponding UDP connection
 * pre: read_template: Read tepmlate set with the TCP sockets that
 *                     have reads pending
 * post: Data is read from the TCP socket for each connection that
 *       had data pending. 
 *       Data is written to corresponding UDP connection
 * return: 0 on success
 *         -1 on error (Actually errors are internally ignored in
 *                      the current implementation)
 **********************************************************************/

int connection_table_read_tcp(fd_set *read_template)
{
        int con;
        struct packet packet;

        bzero(&packet, sizeof(packet));

        for(con=0;con<MAX_CONNECTION;con++){
                if(connection_table_get_status(con)==STATUS_PURGED){
                        continue;
                }
                if(FD_ISSET(connection_table_get_tcp_fd(con), read_template)){
                        if(connection_read_tcp(&(con_table[con]))) {
                                /* Log error but ignore */
                                fprintf(stderr, "connection_table_read_tcp: "
                                                "connection_read_tcp\n");
                        }
                }
        }

        return(0);
}


/**********************************************************************
 * connection_table_set_fd_start
 * Reset internal state about set fds
 * You must call this before the first time you call
 * connection_table_set_fd_next after the read_template being refered
 * changes.
 * pre: none
 * post: Internal state about set fds is reset
 * return: none
 **********************************************************************/

static int __connection_table_next_fd_marker=0;

void connection_table_set_fd_start(void)
{
        __connection_table_next_fd_marker=0;
}


/**********************************************************************
 * connection_table_set_fd_next
 * Find the next connection whose associated tcp socket is in the
 * read_tepmlate
 * pre: read_template: read_template to reference
 * post: internal status of what connection we are up to is updated.
 * return: Next connection whose TCP socket is set in the read_template
 *         -1 if no connection is found
 **********************************************************************/

int connection_table_set_fd_next(fd_set *read_template)
{
        int con;
        int fd;
        int status;
        
        for(con=__connection_table_next_fd_marker;con<MAX_CONNECTION;con++){
                fd = connection_table_get_tcp_fd(con);
                status = connection_table_get_status(con);
                if(status==STATUS_PURGED || status==STATUS_CLOSED || fd<0)
                        continue;
                if(FD_ISSET(fd, read_template)){
                        __connection_table_next_fd_marker=con+1;
                        return(con);
                }
        }
        
        return(-1);
}

/**********************************************************************
 * connection_table_set_fd_end
 * Reset internal state about set fds
 * You should call this after the last time you call
 * connection_table_set_fd_next after the read_template being refered
 * changes.
 *
 * Actually it does nothing, but is here for completeness
 *
 * pre: none
 * post: none
 * return: none
 **********************************************************************/

void connection_table_set_fd_end(void){ ; }



/**********************************************************************
 **********************************************************************
 * The functions that follow enable access by connection
 * number to their connection_table_* counterparts.
 *
 * So connection_table_resend_XXX is a wrapper for
 * connection_table_XXX.
 *
 * Please see connection.h for information on the esmantics of each
 * function
 **********************************************************************
 **********************************************************************/


int connection_table_resend_packet(int con)
{
        extern int iproxy_verbose;

        if(iproxy_verbose > 1)
                printf("resending packet for connection %d\n", con);

        if(connection_resend_packet(&(con_table[con]))){
                fprintf(stderr, "connection_table_resend_packet: "
                                "connection_resend_packet");
                return(-1);
        }


        return(0);
}


int connection_table_send_reply(int con, struct packet *packet)
{
        extern int iproxy_verbose; 

        if(connection_send_reply(&(con_table[con]), packet)){
                fprintf(stderr, "connection_send_reply: "
                                "connection_send_packet\n");
                return(-1);
        }

        if(iproxy_verbose>1)
                printf("Sent reply of size %d to %d\n", 
                                ntohl(packet->length), con);
        return(0);
}


int connection_table_send_end(int con)
{
        extern int iproxy_verbose;

        if(connection_send_end(&(con_table[con]))){
                fprintf(stderr, "connection_table_send_end: "
                                "connection_send_end\n");
                return(-1);
        }

        if(iproxy_verbose > 0)
                printf("sent reply end to %d\n", con);

        return(0);
}


int connection_table_send_ack(int con)
{
        extern int iproxy_verbose;

        if(connection_send_ack(&(con_table[con]))){
                fprintf(stderr, "connection_table_send_ack: "
                                "connection_send_ack\n");
                return(-1);
        }

        if(iproxy_verbose > 0)
                printf("sent ack to %d\n", con);

        return(0);
}


int connection_table_close(int con)
{
        extern int iproxy_verbose;

        if(iproxy_verbose)
                printf("closing connection %d\n", con);

        if(connection_close(&(con_table[con]))){
                fprintf(stderr, "connection_table_close: connection_close\n");
                return(-1);
        }

        return(0);
}


int connection_table_purge(int con)
{
        extern int iproxy_verbose;

        if(iproxy_verbose)
                printf("purging connection %d\n", con);

        if(connection_purge(&(con_table[con]))){
                fprintf(stderr, "connection_table_purge: connection_purge\n");
                return(-1);
        }

        return(0);
}


int connection_table_timeout(int con)
{
        extern int iproxy_verbose;

        if(iproxy_verbose)
                printf("timing out connection %d\n", con);

        if(connection_timeout(&(con_table[con]))){
                fprintf(stderr, "connection_table_timeout: "
                                "connection_timeout\n");
                return(-1);
        }

        return(0);
}


int connection_table_read_template(fd_set *read_template)
{
        int con;
        int open=0;
        int status;

        for(con=0;con<MAX_CONNECTION;con++){
                status=connection_read_template(&(con_table[con]), 
                                read_template);
                if(status < 0) {
                        fprintf(stderr, "connection_table_read_template: "
                                        "connection_read_template\n");
                        continue;
                }
                open+=status;
        }

        return(open);
}

size_t connection_table_write_all_tcp(int con, void *buf, size_t len)
{
        size_t n=0;
        
        if(con>=MAX_CONNECTION || con<0){
                fprintf(stderr, "connection_write_all_tcp: connection %d "
                                "invalid\n", con);
                return(-1);
        }

        if(len && (n=connection_write_all_tcp(&(con_table[con]), buf, len))<=0){
                fprintf(stderr, "connection_table_write_all_tcp: "
                                "connection_write_all_tcp\n");
        }

        return(n);
}


uint32 connection_table_get_status(int con)
{
        if(con > MAX_CONNECTION){
                return(STATUS_ERR);
        }

        return(connection_get_status(&(con_table[con])));
}


uint32 connection_table_get_tcp_fd(int con)
{
        if(con > MAX_CONNECTION){
                return(-1);
        }

        return(connection_get_tcp_fd(&(con_table[con])));
}


uint32 connection_table_get_in_offset(int con)
{
        if(con > MAX_CONNECTION){
                return(-1);
        }

        return(connection_get_in_offset(&(con_table[con])));
}


uint32 connection_table_get_out_offset(int con)
{
        if(con > MAX_CONNECTION){
                return(-1);
        }

        return(connection_get_out_offset(&(con_table[con])));
}


time_t connection_table_get_expire(int con)
{
        if(con > MAX_CONNECTION){
                return(-1);
        }

        return(connection_get_expire(&(con_table[con])));
}


time_t connection_table_get_resend(int con)
{
        if(con > MAX_CONNECTION){
                return(-1);
        }

        return(connection_get_resend(&(con_table[con])));
}


int connection_table_packet_header(int con, struct packet *packet, uint32 flag, 
                uint32 length, uint32 zero_packet)
{
        if(connection_packet_header(&(con_table[con]), packet, flag, length, 
                                zero_packet)<0){
                fprintf(stderr, "connection_table_packet_header: "
                                "connection_packet_header\n");
                return(-1);
        }

        return(0);
}
