/* -*- c-file-style: "linux"; -*- */
/**********************************************************************
 * iproxy-server.c                                            July 2001
 * Horms                                             horms@vergenet.net
 * Andrew Tridgell                                     tridge@samba.org
 *
 * iproxy
 * UDP - TCP proxy
 * Copyright (C) 2001  Andrew Tridgell <tridge@samba.org>
 *                     Horms <horms@vergenet.net>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
 * 02111-1307  USA
 *
 **********************************************************************/

#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <time.h>

#include "util.h"
#include "iproxy.h"
#include "packet.h"
#include "connection.h"
#include "connection_table.h"
#include "interfaces.h"
#include "interface.h"

static unsigned short iproxy_port = PROXY_PORT;
static uint32 iproxy_sid = PROXY_SID;
static unsigned short dest_port = WEB_PORT;
uint32 mcast_ttl = TTL_MULTICAST;
static char *xcast_ip = ADDR_BROADCAST;
static char *dest_ip = ADDR_LOCALHOST;
unsigned short iproxy_mode = MODE_BROADCAST|MODE_SERVER;
int iproxy_verbose;

#define SELECT_TIMEOUT 2

static void do_proxy_server()
{
	struct packet packet;
	size_t n;
	struct sockaddr_in saddr;
        struct timeval tv;
	socklen_t socklen;
        fd_set read_template;
        uint32 offset;
        int status;
        int con;
        int fd;

	while (1) {
                  FD_ZERO(&read_template);

                  interface_add_to_read_template(&read_template);
                  if(connection_table_read_template(&read_template)<0){
                          fprintf(stderr, "do_proxy_server: connection_table_timeout_or_add_to_read_template, ignoring\n");
                  }

                  tv.tv_sec=SELECT_TIMEOUT;
                  tv.tv_usec=0;

                  status=select(FD_SETSIZE, &read_template, NULL, NULL, &tv);
                  if(status<0){
                    if(errno!=EINTR){
                      perror("do_proxy: select");
                      exit(-1);
                    }
                    continue; /* Ignore EINTR */
                  }
                  else if(status==0){
                    continue; /* Skip to start of loop on timeout */
                  }

                  interface_set_fd_start();
                  while((fd=interface_set_fd_next(&read_template))>=0){
                    /* Read packet from UDP socket */
                    bzero(&packet, sizeof(packet));
                    socklen = sizeof(saddr);
                    n = recvfrom(fd, &packet, sizeof(packet),
			     0, (struct sockaddr *)&saddr, &socklen);
                    if(iproxy_verbose)
                        printf("received request of size %d from sid=0x%x "
                                        "cid=0x%x\n", n, ntohl(packet.sid), 
                                        ntohl(packet.cid));

                    if (packet.version != htonl(IPROXY_PROTO_VERSION)){
                            if(iproxy_verbose)
                                printf("[%d] packet for unknown iproxy "
                                                "version\n", getpid());
                            continue;
                    }

                    if(htonl(packet.checksum) != packet_checksum(&packet)){
                            printf("checksum missmatch 0x%x!=0x%x\n",
                                            ntohl(packet.checksum),
                                            packet_checksum(&packet));
                            continue;
                    }

                    if (ntohl(packet.length) != n) continue;
                    if (ntohl(packet.sid) != IPROXY_LIST_SID &&
                            ntohl(packet.sid) != iproxy_sid) continue;

                    con=connection_table_get_connection(dest_ip,
                            dest_port, &saddr, &packet, n);
                    if(con==MAX_CONNECTION){
                            break; /* Error */
                    }
                    if(con>MAX_CONNECTION){
                            continue; /* Control packet */
                    }
                    
                    if (ntohl(packet.sid) == IPROXY_LIST_SID) {
                      n = snprintf(packet.buf, sizeof(packet.buf),
                        "Hello from %d!\n", iproxy_sid);
                      packet.length = htonl(n + PACKET_HDR_LEN);
                      connection_table_send_reply(con, &packet);
                      connection_table_close(con);
                      continue;
                    }

                    connection_table_send_ack(con);

                    offset=connection_table_get_in_offset(con);
                    if (ntohl(packet.offset) != offset) {
                            printf("out of order packet! %d!=%d\n", 
                                            ntohl(packet.offset), offset);
                            continue;
                    }
                          
                    /* Send outgoing packet to TCP socket */
                    connection_table_write_all_tcp(con, packet.buf, 
                                             n-PACKET_HDR_LEN);
                    if(iproxy_verbose > 1)
                        printf("sent request of size %d\n", n-PACKET_HDR_LEN);
                  }
                  interface_set_fd_end();

                  connection_table_set_fd_start();
                  while((con=connection_table_set_fd_next(&read_template))>=0){
                        if(connection_table_get_status(con)==STATUS_UNACKED){
                                printf("Waiting for ack, not reading "
                                       "packet %d\n", con);
                                continue;
                        }
                        n=read(connection_table_get_tcp_fd(con), packet.buf, 
                                        sizeof(packet.buf));
                        if(n<=0){
                                if(connection_table_close(con)){
                                        fprintf(stderr, "do_proxy: "
                                                        "connection_close\n");
                                        exit(-1);
                                }
                                break;
                        }
                        
                        if(iproxy_verbose > 1)
                                printf("received reply of size %d\n", n);
                        if(connection_table_packet_header(con, &packet, 0, 
                                                n+PACKET_HDR_LEN, 0)) {
                                fprintf(stderr, "do_proxy: "
                                                "connection_packet_header\n");
                                connection_table_close(con);
                                continue;
                        }
                        if(connection_table_send_reply(con, &packet)) {
                                fprintf(stderr, "do_proxy: "
                                        "connection_send_reply\n");
                                connection_table_close(con);
                                continue;
                        }
                }
                connection_table_set_fd_end();
        }
}

static void usage(int exit_status)
{
	fprintf(exit_status?stderr:stdout, "
iproxy-server [options]
    -h            help
    -p PORT       listen port number (UDP)
    -d PORT       dest port (TCP)
    -s SID        server id number
    -I IP         set IP to send packets to
    -T TTL        ttl for multicast packets
    -b            Use broadcast (default)
    -m            Use multicast
    -u            Use unicast
    -v            increase verbosity
");

	exit(exit_status);
}

static void dumper(int sig){
        signal(sig, (void(*)(int))dumper);
        connection_table_dump();
}


int main(int argc, char *argv[])
{
	int c;
	char *p;
        char *new_dest_ip=NULL;
        int addr_set=0;

	while ((c = getopt(argc, argv, "s:I:t:bmup:d:vh")) != -1) {
		switch (c) {
		case 's':
			iproxy_sid = atoi(optarg);
			break;
                case 'I':
                        xcast_ip = optarg;
                        addr_set++;
                        break;
                case 't':
                        mcast_ttl = atoi(optarg);
                        break;
                case 'b':
                        iproxy_mode = MODE_BROADCAST|MODE_SERVER;
                        if(!addr_set)
                                xcast_ip = ADDR_BROADCAST;
                        break;
                case 'm':
                        iproxy_mode = MODE_MULTICAST|MODE_SERVER;
                        if(!addr_set)
                                xcast_ip = ADDR_MULTICAST;
                        break;
                case 'u':
                        iproxy_mode = MODE_UNICAST|MODE_SERVER;
                        if(!addr_set)
                                xcast_ip = ADDR_ANY;
                        break;
		case 'p':
			iproxy_port = atoi(optarg);
			break;
		case 'd':
			if ((p = strchr(optarg,':'))) {
				*p = 0;
				dest_port = atoi(p+1);
				new_dest_ip = optarg;
			} else {
				dest_port = atoi(optarg);
			}
			break;
		case 'v':
			iproxy_verbose++;
			break;
		case 'h':
			usage(0);
                        break;
		default:
			usage(-1);
			break;
		}
		
	}

        if(connection_table_init(xcast_ip)) {
                fprintf(stderr, "main: connection_table_init\n");
                return(-1);
        }
        signal(SIGUSR1, (void(*)(int))dumper);

        if(interfaces_get()<0){
                exit(1);
        }

        /* We can just pass the mcast_ip as this
         * argument isn't used unless we are in multicast mode
         */
        if(interface_open_fd(ADDR_ANY, xcast_ip, iproxy_port)){
                fprintf(stderr, "do_proxy: interface_open_fd");
                exit(-1);
        }

        do_proxy_server();
        return 0;
}
